#!/usr/bin/env python3.10
from .result import Err, Ok, Result, UnwrapError

"""New builtins for Lindworm."""

from functools import reduce as fold_right
from itertools import (chain, combinations, count, groupby, permutations,
                       product)

from more_itertools import (adjacent, always_iterable, always_reversible,
                            bucket, circular_shifts, collapse,
                            combination_index, consume, consumer, convolve,
                            countable, distinct_combinations,
                            distinct_permutations, duplicates_everseen,
                            duplicates_justseen, first_true, flatten, ilen,
                            is_sorted, iter_except, iterate, locate, lstrip,
                            minmax, nth_combination, nth_or_last,
                            nth_permutation, nth_product, numeric_range,
                            partitions, permutation_index, powerset,
                            product_index, repeat_last, replace, rlocate,
                            rstrip, set_partitions, sliding_window,
                            split_after, split_at, split_before, split_into,
                            split_when, spy, strip, unique_everseen,
                            unique_in_window, unique_justseen, unique_to_each,
                            with_iter)


def fold_left(*args):
    """Left fold."""
    return fold_right(args[0], reversed(args[1]), *args[2:])



import operator as op

import more_itertools

from .functions import *

OPERATORS = {
    "+": op.add,
    "in": op.contains,
    "/": op.truediv,
    "//": op.floordiv,
    "&": op.and_,
    "^": op.xor,
    "~": op.invert,
    "|": op.or_,
    "**": op.pow,
    "is": op.is_,
    "<<": op.lshift,
    "%": op.mod,
    "*": op.mul,
    "@": op.matmul,
    "not": op.not_,
    ">>": op.rshift,
    "-": op.sub,
    "<": op.lt,
    "<=": op.le,
    "==": op.eq,
    "!=": op.ne,
    ">=": op.ge,
    ">": op.gt,
}


# New operators
OPERATORS |= {
    "isnt": op.is_not,
    "::": it.chain,
    "::*": lambda a, b: it.chain(a, *b),
    ":::": lambda a, b: it.chain(more_itertools.always_iterable(a), more_itertools.always_iterable(b)),
    "??": lambda a, b: a if a is not None else b,
    "..>": lambda f, g: compose(f, g, False, 0),
    "..*>": lambda f, g: compose(f, g, False, 1),
    "..**>": lambda f, g: compose(f, g, False, 2),
    "..": lambda f, g: compose(f, g, False, 0),
    "..*": lambda f, g: compose(f, g, False, 1),
    "..**": lambda f, g: compose(f, g, False, 2),
    "<..": lambda f, g: compose(f, g, True, 0),
    "<*..": lambda f, g: compose(f, g, True, 1),
    "<**..": lambda f, g: compose(f, g, True, 2),
    "$": curry,
    "|>": lambda a, b: b(a),
    "|*>": lambda a, b: b(*a),
    "|**>": lambda a, b: b(**a),
    "<|": lambda a, b: a(b),
    "<*|": lambda a, b: a(*b),
    "<**|": lambda a, b: a(**b),
    "over": map,
    "contains": lambda a, b: b in a
}



import dataclasses as dc
import io
import tokenize
import typing as t

from .constants import constants


@dc.dataclass(eq=True, unsafe_hash=True)
class SimpleToken:
    exact_type: int
    string: str
    indent: int

    def __repr__(self):
        return f"<{type(self).__qualname__} {self.string!r} {constants.tok_name[self.exact_type]} (indent={self.indent})>"

    @property
    def type(self):
        return self.exact_type

    @classmethod
    def tokenize(cls, code: bytes, **kwargs) -> t.Generator:
        """Turns code into a stream of SimpleTokens."""
        stream = tokenize.tokenize(io.BytesIO(code).readline)
        item = cls.from_incomplete_stream(stream, **kwargs)
        return item

    @classmethod
    def from_incomplete_stream(cls, stream: t.Iterable, *, indent=0) -> t.Generator:
        """Turns a tokenize.tokenize stream into a SimpleToken stream, adding whitespace tokens and removing INDENT and DEDENT tokens."""
        INDENT_SIZE = 4
        prev_item = None
        output = []

        for item in stream:
            # Skip indents and dedents and store it in each token instead.
            diff_indent = (item.exact_type == tokenize.INDENT) - \
                (item.exact_type == tokenize.DEDENT)
            indent += diff_indent
            if diff_indent:
                if output:
                    output[-1].indent += diff_indent
                continue

            # Skip encoding and end markers.
            if item.exact_type in {tokenize.ENCODING, tokenize.ENDMARKER}:
                continue

            # Adds whitespace.
            if prev_item and prev_item != item:
                newlines = (item.start[0] - prev_item.end[0])
                spaces = (item.start[1] - prev_item.end[1])

                if spaces and not newlines:
                    output.append(cls(constants.WHITESPACE, " " * spaces, indent))

            result = cls(constants.from_tokenize(item.exact_type), item.string, indent)
            output.append(result)
            prev_item = item

        return output



import dataclasses as dc
import datetime as dt
import io
import itertools as it
import pathlib as p
import sys
import tokenize
import typing as t

import cson
import mako
import mako.template

from .constants import constants
from .rule import Rule
from .token import SimpleToken


@dc.dataclass
class PythonDialectTokenization():
    """A tokenization of a file written in a Python dialect."""
    tokens: t.List[SimpleToken]
    source: str

    # Workaround so that this moves along with stdout
    logger: io.StringIO = dc.field(default_factory=lambda: sys.stdout)

    def to_python(self) -> str:
        return NotImplemented

    @classmethod
    def from_path(cls, path: p.Path, *args, **kwargs):
        with open(path, "rb") as file:
            return cls.from_bytes(file.read(), *args, **kwargs)

    @classmethod
    def from_bytes(cls, item: bytes, *args, **kwargs):
        tokens = SimpleToken.tokenize(item)
        item = cls(list(tokens), item.decode("utf-8"), *args, **kwargs)
        item.rebuild_source(0)
        return item

    @classmethod
    def load_rules(cls) -> list:
        subitems = {}
        order = None

        for filename in cls.rules_dir.iterdir():
            if filename.suffix != ".cson":
                raise ValueError(f"Unexpected file '{filename}' in rules directory '{cls.rules_dir}' (expected extension '.cson'.)")

            if filename.stem == "__order__":
                with open(filename) as file:
                    order = cson.load(file)
                continue

            with open(filename) as file:
                subitems[filename.stem] = cson.load(file)

        # Finds all rules.
        rules = []
        for file_stem in order:
            item = subitems[file_stem]
            rules += [
                Rule(
                    k,
                    **v,
                    priority=constants.priority_of(constants.LPAR)  # -i
                )
                for k, v in item.items()
            ]

        return rules

    def rebuild_source(self, indent=0):
        """Rebuilds the source attribute from tokens."""
        self.source = ""

        for token in self.tokens:
            self.source += token.string

            # Add indentation.
            if token.exact_type in {constants.NEWLINE, constants.NL}:
                self.source += (" " * indent * token.indent)

        return self

    def __iter__(self):
        return iter(self.tokens)



#!/usr/bin/env python3.10
""""""
import pathlib as p


def tempfile():
    """Gets a temporary unused filename."""
    base_path = p.Path("/tmp/")
    filename = None
    i = 0

    while True:
        i += 1
        filename = base_path / f"tempfile_{i}"
        if not filename.exists():
            break

    return filename



import dataclasses as dc
import itertools as it
import typing as t

import more_itertools as mit
import regex as re

ReplacerString = t.TypeVar("RegexReplacementStr", str, bytes)


CASE_SHORTHANDS = {
    "CF": str.casefold,
    "C": str.capitalize,
    "L": str.lower,
    "T": str.title,
    "U": str.upper,
    "S": str.swapcase,
}

MISC_SHORTHANDS = {
    "ju-l": str.ljust,
    "ju-r": str.rjust,
    "st-l": str.lstrip,
    "st-r": str.rstrip,
    "rm-p": str.removeprefix,
    "rm-s": str.removesuffix,
    "cen": str.center,
    "exp": str.expandtabs,
    "rep": str.replace,
    "st": str.strip,
    "tra": str.translate,
    "zfi": str.zfill,
    "len": lambda s: str(len(s)),
    "??": lambda s, r: r if s else s,
}

SHORTHANDS = {**CASE_SHORTHANDS, **MISC_SHORTHANDS}

FORMAT_REGEX = re.compile(r"\{([^{}:]*):(\s*-?[^{}\-:]*)(?>->([^{}]*))?\}")
SUBFORMAT_REGEX = re.compile(r"(\s*\[[^][]+\])|([^:]+)\s*:?")
SUBSUBFORMAT_REGEX = re.compile(r"(\S+)")


# ===| Functions |===


@dc.dataclass
class Replacer:
    regex: re.Regex
    fmtstr: ReplacerString

    def format(self, arg: str, args=None) -> str:
        return self.format_match(re.match(self.regex, arg), args)

    def format_match(self, match: re.match, args=None) -> str:
        return advanced_format(self.fmtstr, match, args)

    # @classmethod
    # def single_formatter(cls, formatter, token_string):
    #     match = re.match(FORMAT_REGEX, formatter)
    #     assert match
    #
    #     # magic: replaces the index with 0
    #     formatter = formatter.replace(match.group(1), "group")
    #     formatter = formatter.replace(match.group(2), "0")
    #     formatter = cls("(.*)", formatter)
    #     return formatter.format(token_string)


def advanced_format(fmtstr: ReplacerString, item: re.Match, subargs=None) -> str:
    # Logging
    print("item:   ", item)
    print("fmtstr:  ", fmtstr)

    subargs = {} if subargs is None else subargs

    if item is None:
        raise ValueError("Expected a match but got None.")

    item_groups = {str(i): v for i, v in enumerate(item.groups())}
    item_groups |= item.groupdict()

    # does not account for double backslashing, but at least it allows escape sequences here
    unescaped_fmtstr = fmtstr.replace(r"\}", "").replace(r"\{", "")
    matches = re.finditer(FORMAT_REGEX, unescaped_fmtstr)
    matches = mit.unique_everseen(matches)

    mapping = {}
    output = fmtstr

    for match in matches:
        origin = match.group(0)

        k = match.group(1).strip()
        lookup_closure = subargs[k] if k != "group" else item_groups
        target = _advanced_format_single(match, fmtstr, item_groups)
        # # Logging
        # print("match:   ", match)
        # print("output:  ", output)
        # print("origin:  ", origin)
        # print("target:  ", target)
        output = output.replace(origin, target)

    return output


def _advanced_format_single(match: re.Match, fmtstr: str, item_groups: dict) -> str:
    """Regex transformation thing."""

    match_groups = [
        str.strip(i) if i is not None else None for i in match.groups()]

    if match_groups[1] not in item_groups:
        raise ValueError(
            f"No group named {match_groups[1]!r} could be found (valid groups: {sorted(item_groups.keys())}).")

    # print(match_groups)
    # print(item_groups)

    output = item_groups[match_groups[1]]
    formatting_descriptor = match_groups[2]

    if formatting_descriptor:
        instructions = re.finditer(SUBFORMAT_REGEX, formatting_descriptor)
        instructions = [i.group(0) for i in instructions]

        instructions = [
            [j.removesuffix(":") for j in re.findall(SUBSUBFORMAT_REGEX, i)] for i in instructions
        ]

        for instruction in instructions:
            instruction = list(filter(None, instruction))
            command = instruction[0].strip()
            arguments = instruction[1:]

            # Calls a method.
            for k, v in SHORTHANDS.items():
                if command == k:
                    command = v
                    print(output)
                    output = v(output, *arguments)
                    print(output)
                    break
            else:
                # Uses normal string formatting.
                assert not arguments
                metaformatter = "{0" + command + "}"

                output = metaformatter.format(
                    _StrSlicer(output)
                )
            continue

    return output


# ===| Utilities |===

@dc.dataclass
class _StrSlicer:
    """Utility class that implements __getitem__ such that it accepts a string version of its normal input.

    Useful for string formatting.

    Examples
    ========
    >>> item = list(range(30))
    >>> _StrSlicer(item)["10: 15"] # Spaces work fine with this as well
    [10, 11, 12, 13, 14]
    >>> "The numbers in range are {0[10:15]}.".format(_StrSlicer(item))
    "The numbers in range are [10, 11, 12, 13, 14]."
    """

    item: t.Any

    def __getitem__(self, i):
        return self.item[_parse_slice(i)]


def _parse_slice(item: str) -> slice:
    """Parses a slice representation into a slice.

    Examples
    ========
    >>> _parse_slice("10:20:-1")
    slice(10, 20, -1)
    """
    # https://stackoverflow.com/a/51105983
    if ":" not in item:
        return int(item)

    return slice(*map(lambda x: int(x.strip()) if x.strip() else None, item.split(':')))



"""Header functions for Lindworm."""
import itertools as it
import typing as t

_IOTA = 0


def iota(zero=False) -> int:
    global _IOTA
    if zero:
        _IOTA = 0
    else:
        _IOTA += 1
    return _IOTA


def curry(fn: t.Callable, *args, **kwargs) -> t.Callable:
    return lambda *args_, **kwargs_: fn(*args, *args_, **kwargs, **kwargs)


def compose(left, right, reverse, num_stars) -> t.Callable:
    """Returns left `of` right, or right `of` left if reversed."""
    if reverse:
        left, right = right, left

    if num_stars == 1:
        return lambda *args, **kwargs: left(*right(*args, **kwargs))
    if num_stars == 2:
        return lambda *args, **kwargs: left(**right(*args, **kwargs))

    return lambda *args, **kwargs: left(right(*args, **kwargs))



import dataclasses as dc
import typing as t

import more_itertools as mit


@dc.dataclass
class LindwormIteratorSlicer:
    """A class for slicing iterators."""
    iterator: t.Iterator

    def __getitem__(self, index):
        if isinstance(index, int):
            for _ in range(index):
                next(self.iterator)
            return next(self.iterator)
        if isinstance(index, slice):
            return mit.islice_extended(self.iterator, index.start, index.stop, index.step)

        raise TypeError(f"{type(self).__qualname__} indices must be integers or slices, not {type(index).__qualname__}")



import regex_replace

Uppercaser = regex_replace.Replacer(
    r"(.*)",
    "{ group : 0 -> U}"
)

assert Uppercaser.format("hello world") == "HELLO WORLD"

Uppercaser = regex_replace.Replacer(
    r"(.*)",
    "{ group : 0 -> U }"
)



"""Token constants."""
import tokenize


_IOTA = 0


def iota():
    global _IOTA
    _IOTA += 1
    return _IOTA


class constants:
    ENDMARKER = iota()
    NAME = iota()
    NUMBER = iota()
    STRING = iota()
    NEWLINE = iota()
    INDENT = iota()
    DEDENT = iota()
    LPAR = iota()
    RPAR = iota()
    LSQB = iota()
    RSQB = iota()
    COLON = iota()
    COMMA = iota()
    SEMI = iota()
    PLUS = iota()
    MINUS = iota()
    STAR = iota()
    SLASH = iota()
    VBAR = iota()
    AMPER = iota()
    LESS = iota()
    GREATER = iota()
    EQUAL = iota()
    DOT = iota()
    PERCENT = iota()
    LBRACE = iota()
    RBRACE = iota()
    EQEQUAL = iota()
    NOTEQUAL = iota()
    LESSEQUAL = iota()
    GREATEREQUAL = iota()
    TILDE = iota()
    CIRCUMFLEX = iota()
    LEFTSHIFT = iota()
    RIGHTSHIFT = iota()
    DOUBLESTAR = iota()
    PLUSEQUAL = iota()
    MINEQUAL = iota()
    STAREQUAL = iota()
    SLASHEQUAL = iota()
    PERCENTEQUAL = iota()
    AMPEREQUAL = iota()
    VBAREQUAL = iota()
    CIRCUMFLEXEQUAL = iota()
    LEFTSHIFTEQUAL = iota()
    RIGHTSHIFTEQUAL = iota()
    DOUBLESTAREQUAL = iota()
    DOUBLESLASH = iota()
    DOUBLESLASHEQUAL = iota()
    AT = iota()
    ATEQUAL = iota()
    RARROW = iota()
    ELLIPSIS = iota()
    COLONEQUAL = iota()
    OP = iota()
    AWAIT = iota()
    ASYNC = iota()
    TYPE_IGNORE = iota()
    TYPE_COMMENT = iota()
    SOFT_KEYWORD = iota()
    ERRORTOKEN = iota()
    COMMENT = iota()
    NL = iota()
    ENCODING = iota()
    WHITESPACE = iota()

    tok_name = {value: name for name, value in locals().items() if isinstance(value, int) and not name.startswith('_')}

    EXACT_TOKEN_TYPES = {
        '!=': NOTEQUAL,
        '%': PERCENT,
        '%=': PERCENTEQUAL,
        '&': AMPER,
        '&=': AMPEREQUAL,
        '(': LPAR,
        ')': RPAR,
        '*': STAR,
        '**': DOUBLESTAR,
        '**=': DOUBLESTAREQUAL,
        '*=': STAREQUAL,
        '+': PLUS,
        '+=': PLUSEQUAL,
        ',': COMMA,
        '-': MINUS,
        '-=': MINEQUAL,
        '->': RARROW,
        '.': DOT,
        '...': ELLIPSIS,
        '/': SLASH,
        '//': DOUBLESLASH,
        '//=': DOUBLESLASHEQUAL,
        '/=': SLASHEQUAL,
        ':': COLON,
        ':=': COLONEQUAL,
        ';': SEMI,
        '<': LESS,
        '<<': LEFTSHIFT,
        '<<=': LEFTSHIFTEQUAL,
        '<=': LESSEQUAL,
        '=': EQUAL,
        '==': EQEQUAL,
        '>': GREATER,
        '>=': GREATEREQUAL,
        '>>': RIGHTSHIFT,
        '>>=': RIGHTSHIFTEQUAL,
        '@': AT,
        '@=': ATEQUAL,
        '[': LSQB,
        ']': RSQB,
        '^': CIRCUMFLEX,
        '^=': CIRCUMFLEXEQUAL,
        '{': LBRACE,
        '|': VBAR,
        '|=': VBAREQUAL,
        '}': RBRACE,
        '~': TILDE,
    }

    # https://docs.python.org/3/reference/expressions.html#operator-precedence
    PRIORITY_DICT = dict(enumerate((
        [LPAR, RPAR, LSQB, RSQB],
        [DOT],
        [DOUBLESTAR, TILDE],
        [PLUS, MINUS, STAR, SLASH, LEFTSHIFT, RIGHTSHIFT],
        [AMPER, CIRCUMFLEX, VBAR],
        [GREATER, GREATEREQUAL, LESS, LESSEQUAL],
    )))

    # ===| Functions |===

    @classmethod
    def from_tokenize(cls, i: int) -> int:
        name = tokenize.tok_name[i]
        rename = {
            "NL": "NEWLINE"
        }
        return getattr(cls, rename.get(name, name))

    @classmethod
    def priority_of(cls, i: int) -> int:
        for k, v in cls.PRIORITY_DICT.items():
            if i in v:
                return k

        return -float("inf")



#!/usr/bin/env python3.10
from .misc import reversed_dict
from .pretty import (format_si, indent, polite_error, polite_success,
                     polite_warning, pretty_table, progressbar_iterator)
from .regex_replace import (CASE_SHORTHANDS, FORMAT_REGEX, MISC_SHORTHANDS,
                            SHORTHANDS, SUBFORMAT_REGEX, SUBSUBFORMAT_REGEX,
                            Replacer, advanced_format)
from .tempfile import tempfile



import dataclasses as dc
import json
import typing as t

import more_itertools as mit
import regex as re

from ..utils import *
from .constants import constants

# FORMAT_REGEX = re.compile(r"\{([^{}:]*):(-?[^{}\-:]*)(?>->([^{}]*))?\}")


@dc.dataclass(eq=True, unsafe_hash=True)
class Rule:
    name: str
    regex: re.Pattern
    formatter: str
    tags: list
    priority: int = constants.priority_of(constants.LPAR)

    def __repr__(self):
        return f"<{type(self).__qualname__} {self.name!r}>"

    def __lt__(self, o):
        return self.priority < o.priority

    def __init__(self, name, regex, tags=None, formatter=None, *, consumed=None, delimiters=None, priority=constants.priority_of(constants.LPAR)):
        self.name = name
        self.regex = re.compile(regex, re.W | re.M | re.X)
        # self.formatter = Replacer(regex, formatter)
        self.formatter = formatter
        self.consumed = consumed
        self.tags = tuple(tags) if tags else tuple()
        self.priority = self.priority

        if self.consumed is None:
            # Automatically find consumed groups from the formatter.
            items = re.finditer(FORMAT_REGEX, formatter)
            items = [i.group(2) for i in items if i.group(1) == "token"]
            formatters = mit.unique_everseen(items)
            self.consumed = list(map(int, formatters))

        self.delimiters = delimiters if delimiters is not None else {}



import dataclasses as dc
import datetime as dt
import hashlib
import io
import itertools as it
import os
import pathlib as p
import subprocess
import sys
import tokenize
import typing as t

import cson
import mako
import mako.template
import more_itertools as mit
import numpy as np
import regex as re

from ..utils import *
from .constants import constants
from .dialect import PythonDialectTokenization
from .rule import FORMAT_REGEX, Rule
from .token import SimpleToken

__version__ = "0.1.0"

# ===| Globals |===

TMP_DIR = p.Path("/tmp")

METADATA_BEGIN = "===| BEGIN LINDWORM METADATA |==="
METADATA_FINAL = "===| END LINDWORM METADATA |==="
PARENT_DIR = p.Path(__file__).parent

BALANCED_TOKENS = {
    constants.LPAR: constants.RPAR,
    constants.LSQB: constants.RSQB,
}

# ===| Classes |===


class LindwormTokenization(PythonDialectTokenization):
    """A tokenization of a file written in the Lindworm Python dialect."""
    template_dir = PARENT_DIR / "../data/templates"
    rules_dir = PARENT_DIR / "../data/rules"

    def generate_metadata(self):
        origin_digest = hashlib.md5((self.source + __version__).encode("utf-8")).hexdigest()
        metadata = {
            "uid": origin_digest,
            "compiler": {"name": "sigurd", "version": __version__},
            "compilation_time": dt.datetime.now().strftime("%m/%d/%Y, %H:%M:%S"),
            "dialect": "lindworm"
        }

        return metadata

    def parse_file_metadata(self, filename: str) -> str:
        """Parse file metadata from a filename. Returns None if the metadata is invalid."""
        with open(filename) as file:
            return self.parse_metadata(file.read())

    def parse_metadata(self, string: str) -> t.Optional[dict]:
        """Parse file metadata from string contents. Returns None if the metadata is invalid."""

        if not ("# " + METADATA_BEGIN in string and "# " + METADATA_FINAL in string):
            return None

        metadata_block = string.split("# " + METADATA_BEGIN)[1].split("# " + METADATA_FINAL)[0]
        metadata_block = (i.removeprefix("# ") for i in metadata_block.splitlines(True))
        try:
            return cson.loads("".join(metadata_block))
        except cson.ParseError:
            return None

    def to_python(self, *, beautifier="none"):
        self.logger.write(f"INFO | Compiling self.\n")
        self.convert_multiline_comments()

        metadata = self.generate_metadata()
        self.compile_tokens()
        template = load_template(self.template_dir / "default.py.mako")

        metadata_lines = cson.dumps(metadata, indent=4).splitlines()
        metadata_lines.insert(0, METADATA_BEGIN)
        metadata_lines.append(METADATA_FINAL)

        metadata_str = "\n".join(map("# ".__add__, metadata_lines))

        compiled = template.render(
            source=self.source,
            metadata=metadata_str
        )

        if beautifier == "autopep8":
            import autopep8
            compiled = autopep8.fix_code(compiled, {'aggressive': 3})

        elif beautifier == "black":
            # Buffer the code into a temporary file to run black
            # from the command line on it.
            filename = tempfile()
            with open(filename, "w", encoding="utf-8") as file:
                file.write(compiled)
            subprocess.run(["black", filename, "-q"])
            with open(filename, "r", encoding="utf-8") as file:
                compiled = file.read()

        elif beautifier != "none":
            raise ValueError(f"Unknown beautifier '{beautifier}'")

        self.logger.write(f"INFO | Compilation done!.\n")

        return compiled

    def compile_tokens(self):
        pointer = 0
        rules = self.load_rules()
        ignored = {constants.STRING, constants.COMMENT}

        matches_table = []
        maxloop = 128
        maxloop_inner = None

        # TODO: get memoization working right with returning the affecte ranges

        while rules and maxloop:
            maxloop_inner = 20
            maxloop -= 1

            # Try to find a new target to search and replace.
            while rules and not matches_table:
                maxloop_inner -= 1
                if not maxloop_inner:
                    raise ValueError("Stuck in infloop")
                pointer = 0
                first_matches = []
                self.logger.write(f"RULES | {rules}\n")

                # Get matches for one rule at a time.
                for rule in rules.copy():
                    for result in re.finditer(rule.regex, self.source):
                        first_matches.append((rule, result))

                    first_matches = [
                        (k, v) for k, v in first_matches
                        if v is not None
                        and self.exact_type_at(v.span()[0]) not in ignored
                    ]

                    first_matches = sorted(
                        first_matches,
                        key=lambda x: (x[1].span()[0], - x[1].span()[1]),
                        reverse=True
                    )

                    if first_matches:
                        break
                    rules.pop(0)

                matches_table = first_matches[:1]

            # Since no new ones were found, we're done!
            if not rules:
                break

            match = matches_table.pop(-1)
            pointer = match[1].span()[0]

            # Remove matches that will get moved around due to the overlapping windows.
            matches_table = [(k, v) for k, v in matches_table if v.span()[1] < pointer]
            self.logger.write(f"MATCHESTABLE | {matches_table}\n")
            self.logger.write(f"MATCH | {match}\n")
            # self.logger.write(repr(match) + "\n" + repr(match.groups()))

            # Pointers on either side of the match.
            low = self.source2token_idx(match[1].span()[0])
            high = self.source2token_idx(match[1].span()[1])

            self.logger.write(f"RULE | Match found between {(range(low, high))}\n")
            rule = match[0]
            mapping, consumed = self.get_replacement(rule, match, range(low, high))
            self.logger.write(f"RULE | {mapping, consumed}\n")
            self.logger.write(f"TOKENS | {min(consumed)} and up: {self.tokens[min(consumed):max(consumed)]}\n")

            new_item = rule.formatter
            for key, value in mapping.items():
                if key not in new_item:
                    raise ValueError("Something's very odd here")
                new_item = new_item.replace(key, value)

            self.logger.write(f"RULE | Formatted {rule.formatter!r} -> {new_item!r}\n")

            # Removes the old items and adds the replacements.
            token_dict = dict(enumerate(self.tokens))
            token_dict = {k: [v] for k, v in token_dict.items() if k not in consumed}

            tokenization_of_subgroup = list(SimpleToken.tokenize(new_item.encode("utf-8")))
            token_dict[low] = tokenization_of_subgroup

            self.tokens = [i[1] for i in sorted(token_dict.items())]
            self.tokens = list(mit.flatten(self.tokens))
            self.rebuild_source()

            self.logger.write(f"SOURCE | {self.source!r}\n")

        else:
            raise ValueError("Entered infinite loop.")

        self.rebuild_source(4)
        return self

    def neighbouring_tokengroup(self, pointer: int, direction: int, rule) -> dict:
        """Returns the neighbouring group of tokens."""
        enumerated = list(enumerate(self.tokens))

        if direction == 1:
            token_items = enumerated[pointer:][::direction]
            balancer = reversed_dict(BALANCED_TOKENS.copy())
        else:
            token_items = enumerated[:pointer][::direction]
            balancer = BALANCED_TOKENS.copy()

        done = False
        force_terminate = False
        output = []
        consumed = set()
        balanced_tokens_depths = {k: 0 for k in balancer}

        ENDING = {constants.WHITESPACE, constants.ERRORTOKEN, constants.NUMBER, constants.STRING,
                  constants.DOT, constants.NAME, constants.COLON, constants.LBRACE, constants.RBRACE}
        LEADING = {constants.WHITESPACE, constants.ERRORTOKEN, constants.NEWLINE}
        TERM = {constants.EQUAL}

        for index, token in token_items:
            self.logger.write(f"TOKEN | {token} {done} {direction}\n")

            if (
                (
                    # Avoid going into a new line if the line is balanced.
                    token.exact_type == constants.NEWLINE
                    and not any(balanced_tokens_depths.values())
                )
                or (
                    # Avoid going backwards into assignments.
                    token.exact_type in TERM
                    and direction == -1
                )
                or (
                    token.exact_type in balanced_tokens_depths.keys()
                    and not any(balanced_tokens_depths.values())
                )
            ):
                break

            # Avoids returning early from consuming leading whitespace tokens.
            if (
                (not done and token.exact_type in LEADING)
                or token.exact_type in ENDING
            ):
                output.append(index)
                consumed.add(index)
                continue

            # Ensures balanced tokens.
            if (
                done
                and not any(balanced_tokens_depths.values())
                and (
                    # If this check is not done, the program will attempt to grab
                    # left parentheses as well: for example, "greeter(greeting"
                    # when the expected result is "greeting".
                    direction == -1
                    or constants.priority_of(token.exact_type) < rule.priority
                )
            ):
                break

            # Add or subtract to the balance if necessary.
            for k, v in balancer.items():
                if token.exact_type in (k, v):
                    balanced_tokens_depths[k] += ((token.exact_type == k) - (token.exact_type == v))
                    break

            if token.exact_type != constants.INDENT:
                output.append(index)
                consumed.add(index)
            done = True

        return {
            "consumed": consumed,
            "output": output[::direction]
        }

    def get_replacement(self, rule: Rule, match, matched_range: list) -> tuple:
        items = re.finditer(FORMAT_REGEX, rule.formatter)
        items = mit.unique_everseen(items, key=lambda x: x.groups())
        items = list(items)
        mapping = {}
        consumed = set()
        self.logger.write(f"SUBMATCHES | {items}\n")

        bucket_items = mit.bucket(items, lambda x: x.group(1).strip())

        mapping = {}
        maxtoken, mintoken = -np.inf, np.inf
        token_set = set()
        sometokenstuff = {}
        token_mapping = {}

        # Group format: Use a named capture group.
        for token_submatch in bucket_items["group"]:
            token_submatch = token_submatch.group()
            replacer = Replacer(rule.regex, token_submatch)
            mapping[token_submatch] = replacer.format_match(match[1])

        # Token format: Get the nth token relative to this one.

        # Finds the maximum and minimum required token.
        for token_submatch in bucket_items["token"]:
            num = int(token_submatch.group(2))
            maxtoken = max(maxtoken, num)
            mintoken = min(mintoken, num)
            token_set.add(num)
            sometokenstuff[num] = sometokenstuff.get(num, [])
            sometokenstuff[num].append(token_submatch)

        # Gets all items between the maximum and minimum token.
        for submatch in {maxtoken, mintoken} - {-np.inf, np.inf}:
            direction = np.sign(submatch)
            inner_pointer = matched_range.stop if direction == 1 else matched_range.start

            for n in range(np.abs(submatch)):
                token_indices = self.neighbouring_tokengroup(inner_pointer, direction, rule)
                self.logger.write((f"TOKENINDICES | {token_indices!r}\n"))
                assert token_indices["consumed"]
                inner_pointer = max(token_indices["consumed"]) if direction == 1 else min(token_indices["consumed"])

                matched_items = [self.tokens[i].string for i in token_indices["output"]]
                matched_items_string = "".join(matched_items)

                # Find all of the token subgroups that were based on this particular match.
                # Then processes them in a for loop since they might have different formattings.
                if (direction * (n + 1)) in sometokenstuff:
                    for token_submatch in sometokenstuff[direction * (n + 1)]:

                        token_name = token_submatch.group()

                        # MAGIC
                        # HACK: Should change internal class mechanics to be able to operate on strings as well.
                        temp = token_name.replace("token", "group")
                        temp = temp.replace(re.search(r"-?\d+", temp).group(), "0")
                        # MAGIC

                        replacer = Replacer("(.*)", temp)
                        mapping[token_name] = replacer.format(matched_items_string)

                # what does this do
                token_mapping[direction * (n + 1)] = token_indices

        # Only consume the items that should be.
        for key in token_set:
            consumed |= set(token_mapping[key]["output"])
            consumed |= token_mapping[key]["consumed"]

        consumed |= set(matched_range)
        return mapping, consumed

    def _token_range(self, i: int) -> range:
        # Sums the lengths of the previous lines and adds the line location.
        start = sum(map(lambda x: len(x.string), self.tokens[:i]))
        out = range(start, start + len(self.tokens[i].string))

        return out

    def source2token_idx(self, idx: int) -> int:
        """Returns the index of the source item which contains the string index in the source."""
        for i in range(len(self.tokens)):
            if idx < self._token_range(i).stop:
                if idx in self._token_range(i):
                    return i
                return None

        raise ValueError(f"Index {idx} out of matched_range.")

    def exact_type_at(self, index: int) -> int:
        """Gets the type at a location."""
        idx = self.source2token_idx(index)
        return self.tokens[idx].exact_type

    def convert_multiline_comments(self):
        """Removes multiline comments denoted by ###-blocks from the token stream."""
        in_comment = False
        new_tokens = []

        # TODO: This just removes them but it would be nice to keep them
        for token in self.tokens:
            if token.exact_type == constants.COMMENT and token.string.startswith("###"):
                in_comment = not in_comment
                continue
            if in_comment:
                continue
            new_tokens.append(token)

        self.tokens = new_tokens
        self.rebuild_source()

# ===| Functions |===


def load_template(path: p.Path) -> mako.template:
    with open(path) as file:
        contents = file.read()
    template = mako.template.Template(contents)
    return template


def parse_metadata(contents):
    md_begin_index = contents.index(METADATA_BEGIN)
    md_final_index = contents.index(METADATA_FINAL)
    return contents[:md_begin_index] + contents[md_final_index:], contents[md_begin_index:md_final_index]



#!/usr/bin/env python3.10
from .constants import constants
from .dialect import PythonDialectTokenization
from .lindworm import LindwormTokenization
from .rule import Rule
from .token import SimpleToken



#!/usr/bin/env python3.10
import typing as t

T, V, E, Q, W = map(t.TypeVar, "TVEQW")
OriginalSignature = t.Callable[Q, W]


class UnwrapError(BaseException):
    """An error raised by the Result type on failed unwrappings."""


class Result(t.Generic[T]):
    def __init__(self):
        raise NotImplementedError(f"Instances of type `{type(self)}` should not be initialized directly.")

    def __repr__(self) -> str:
        return f"<{type(self).__name__}: {self.contents!r}>"

    def unwrap(self) -> T:
        """Attempt to unwrap the value. Raises UnwrapError on failure."""
        if not self:
            raise UnwrapError
        return self.contents

    def unwrap_err(self) -> T:
        """Attempt to unwrap the value, expecting an Err. Raises UnwrapError on failure."""
        if self:
            raise UnwrapError
        return self.contents

    def expect(self, error_msg: str) -> T:
        """Attempt to unwrap the value. Raises UnwrapError with the error message on failure."""
        if not self:
            raise UnwrapError(error_msg)
        return self.contents

    def expect_err(self, error_msg: str) -> T:
        """Attempt to unwrap the value, expecting an Err. Raises UnwrapError with the error message on failure."""
        if self:
            raise UnwrapError(error_msg)
        return self.contents

    def unwrap_or(self, value: T) -> T:
        """Attempt to unwrap the value. Returns `value` on failure."""
        return self.unwrap() if self else value

    def unwrap_or_else(self, func: t.Callable[[], T]) -> T:
        """Attempt to unwrap the value. Returns `func()` on failure."""
        return self.unwrap() if self else func()

    def is_ok(self) -> bool:
        """Returns a bool corresponding to whether or not the item is Ok."""
        return bool(self)

    def is_err(self) -> bool:
        """Returns a bool corresponding to whether or not the item is Err."""
        return not bool(self)

    def copy(self):
        """Returns a copy of the item."""
        return type(self)(self.contents)

    def __eq__(self, other):
        return (
            self.is_ok() == other.is_ok()
            and self.unwrap() == other.unwrap()
        )

    @classmethod
    def from_iterator(cls, iterator):
        """Turns an iterator of Result instances to a Result containing an iterator.

        If any of the items of the iterator is Err, the function short-circuits and returns this Err.
        Note that this might mean that the iterable is not guaranteed to be exhausted."""
        contents = []
        for item in iterator:
            if item.is_err():
                return item
            if not isinstance(item, Result):
                raise TypeError(f"Result.from_iterator only support Result instances, not {type(item)}")
            contents.append(item.unwrap())

        return Ok(contents)

    # === Decorators ===
    def wrap_with_closure(optional_func: t.Callable[[UnwrapError], t.Any] = (lambda x: ...)) -> t.Callable[OriginalSignature, OriginalSignature]:
        """
        Meta-decorator that calls `optional_func` on the exception and
        returns Err() if the function raises an UnwrapError.
        """
        def inner(func: OriginalSignature) -> OriginalSignature:
            def replacement_func(*args, **kwargs) -> Result[W]:
                try:
                    return Ok(func(*args, **kwargs))
                except UnwrapError as exception:
                    optional_func(exception)
                    return Err(exception)

            return replacement_func

        return inner

    def wrap(func: OriginalSignature) -> OriginalSignature:
        """Decorator that returns Err() if the function raises an UnwrapError."""
        return wrap_with_closure()(func)


class Ok(Result):
    def __init__(self, contents: V):
        self.contents = contents

    def __bool__(self):
        return True

    def __hash__(self):
        return hash((Ok, self.contents))

    def __iter__(self):
        return iter((self.contents,))


class Err(Result):
    def __init__(self, contents: E):
        self.contents = contents

    def __bool__(self):
        return False

    def __hash__(self):
        return hash((Err, self.contents))

    def __iter__(self):
        return iter(tuple())



import pathlib as p

from setuptools import find_namespace_packages, setup

setup(
    name='lindworm-language',
    version='0.1.0',
    description='Lindworm programming language compiler and standard library',
    url='https://github.com/midnattssol/lindworm',
    author='midnattssol',
    author_email='cd154a7e15@protonmail.com',
    license='BSD 2-clause',
    packages=find_namespace_packages(include=['lindworm.*']),
    install_requires=[
        "numpy",
        "autopep8",
        "mako",
        "more_itertools",
        "regex",
        "colorama",
        "frozendict",
        "black"
    ],
    extras_require={
        'autopep8 linting': ["autopep8"],
        'black linting': ["black"],
    },
    scripts=[
        "bin/lindworm",
        "bin/sigurd"
    ],
    classifiers=[
        'Development Status :: 1 - Planning',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
    ],
    include_package_data=True
)



import alive_progress
import io
import sys
import colorama
import numpy as np

colorama.init()


class indent:
    def __init__(self, char=" " * 4, vspace=0):
        self.char = char
        self.buffer = io.StringIO()
        self.vspace = vspace

    def __enter__(self):
        sys.stdout = self.buffer

    def __exit__(self, *args):
        sys.stdout = sys.__stdout__
        for _ in range(self.vspace):
            print()
        self.buffer.seek(0)
        lines = self.buffer.getvalue().splitlines()
        self.buffer.close()
        print("\n".join(map(self.char.__add__, lines)))
        for _ in range(self.vspace):
            print()


def polite_warning(warning: str):
    print(f"{colorama.Fore.YELLOW}Warning: {warning} {colorama.Style.RESET_ALL}")


def polite_success(notice: str):
    print(f"{colorama.Fore.GREEN}Success: {notice} {colorama.Style.RESET_ALL}")


def polite_error(error: str, *, code: int = 1):
    print(f"{colorama.Fore.RED}Fatal: {error} {colorama.Style.RESET_ALL}")
    exit(code)


def pretty_table(table, *, spacer="  ", title=None):
    tb = []
    for column in zip(*table):
        maxwidth = max(map(lambda x: len(str(x)), column))
        tb.append(list(map(lambda x: str(x).ljust(maxwidth), column)))

    if title is not None:
        title += "\n" + len(title) * "=" + "\n"
    else:
        title = ""
    return title + "\n".join(spacer.join(i) for i in zip(*tb))


def progressbar_iterator(it):
    return alive_progress.alive_it(it, theme="smooth")


def format_si(number: float, *, significant_digits: int = 3, long: bool = False):
    """Format a number using SI prefixes.

    The prefix is chosen such that the number before the prefix is 1 < x < 1000."""

    # From https://gist.github.com/ignamv/1dde19d5a15f4562c86fade74ded2f7e.
    if number == 0:
        return '0 '

    if long:
        prefixes = ["atto", "femto", "pico", "nano", "micro", "milli", "", "kilo", "mega", "giga", "tera", "peta", "exa"]
    else:
        prefixes = [*'afpnμm', '', *'kMGTPE']

    multipliers = 10. ** np.arange(-18, 19, 3)
    inv_multipliers = 10. ** (-np.arange(-18, 19, 3))
    abs_number = abs(number)
    index = np.searchsorted(multipliers, abs_number) - 1
    prefix = prefixes[index]
    number *= inv_multipliers[index]
    abs_number = inv_multipliers[index] * abs_number

    if (abs_number > 1000 or abs_number < 1) and (index == -1 or index == len(multipliers) - 1):
        format_ = '{:.3e} '
        number *= multipliers[index]
    else:
        if abs_number < 10:
            digits_before_comma = 1
        elif abs_number < 100:
            digits_before_comma = 2
        else:
            digits_before_comma = 3
        decimals = max(0, significant_digits - digits_before_comma)
        fmt = '{:.' + str(decimals) + 'f} {}'

    return fmt.format(number, prefix)
